      integer function cpft(r, i, n, incp, signp)
c  Value returned is 0 for success, >0 for failure.
c  Can also be called as a subroutine, as before.
c
c  Fortran transliteration of singleton's 6600 assembly-coded fft.
c  Intended to be of assistance in understanding his code, and in
c  future writing of an fft for another machine.
c  It should be translated into machine code rather than used for
c  production as is because it is versatile and efficient enough to
c  see lots of use, it benefits greatly from careful hand coding, and is
c  short and simple enough to do quickly.
c  A. Bruce Langdon, LRL, 1971.
c  Revised 1985, 1990.
c
c  Comments below are mostly from 6600-7600 version.
c
c  r	  real part of data vector.
c  i	  imag part of data vector.
c  n	  number of elements (=1,2,4,8...).
c  inc	  spacing in memory of data (usually 1, but see below).
c  sign	  its sign will be sign of argument in transform exponential.
c
c    On entry arrays r and i contain the sequence to be transformed.
c  On exit they contain the transform.  Input and output sequences are
c  both in natural order (i.e. not bit-reversed scrambled).
c
c    A call to cpft with sign=+1, followed by another call with the
c  first 4 parameters the same and sign=-1, will leave r and i with
c  their original contents times n.  The same is true if first sign=-1,
c  and next sign=+1.
c
c    The usefulness of parameter inc may be illustrated by 2 examples:
c    Suppose the complex sequence is stored as a fortran complex array
c  z, i.e. real and imaginary parts in alternate memory cells.  The
c  separation between consecutive real (or imaginary) elements is 2
c  words, so inc=2.  The call might be
c	   call cpft(real(z), aimag(z), n, 2, sign)
c  For many compilers one would instead have to do something like
c	   call cpft(ri, ri(2), n, 2, sign)
c  where ri is a real array equivalenced to z.
c    Suppose one had an array c with dimensions n1, n2.  One wants r to
c  be row i1 and i to be row i2.  The separation of consecutive elements
c  is n1 and starting addresses are c(i1,1) and c(i2,1), so use
c	   call cpft(c(i1,1), c(i2,1), n2, n1, sign)
c
c  Timing, assuming minimal memory bank conflicts:
c    6400 time for n=1024 is 220,000=21.5*n*log2n microseconds.
c    6600 time for n=1024 is  44,500=4.35*n*log2n microseconds.
c    7600 time for n=1024 is   8,300=0.81*n*log2n microseconds.
c
c    A radix 2 fft provokes memory bank conflicts at best, but timing
c  is noticeably worsened when like elements of r and i are in the same
c  bank and/or inc is a multiple of a power of 2.  In a worst case on
c  the 7600 the speed was decreased by a factor of 3.
c    Thus in the example above, if n1=multiple of 32 one might
c  decide to waste a little memory by increasing n1 to 33, thus
c  decreasing conflicts for transforms over rows or over columns.
c
c  Written by R. C. Singleton, Stanford Research Institute, Nov. 1968.
c  Commentary, LRL linkage and other minor changes by A. Bruce Langdon
c  Lawrence Radiation Laboratory, Livermore, April 1971.
c
c  References:
c    (1) R. C. Singleton, 'On computing the fast Fourier transform',
c	 Comm. Assoc. Comp. Mach. vol. 10, pp. 647-654 (1967).
c    (2) R. C. Singleton, Algorithm 345 'An Algol convolution procedure
c	 based on the fast Fourier transform', Comm. ACM vol. 12,
c	 pp. 179-184 (1969).
c    (3) W. M. Gentleman and G. Sande, 'Fast Fourier transforms - for
c	 fun and profit', Proc. AFIPS 1966 fall joint computer conf.,
c	 vol. 29, pp. 563-578.
c
      real r(0:1), i(0:1)
c  Length n can be as large as 2**log2nmax.
      parameter (log2nmax=20)
      integer signp, span, rc
      real sines(log2nmax), i0, i1
      save sines
c
c  Table of sines.
c    These should be good to the very last bit.  They are given in octal
c  to prevent an assembler from converting them poorly.  They may be
c  obtained by evaluating the indicated sines in double precision and
c  writing them out in octal format (a single precision sine routine
c  is not accurate enough).  Use the most significant word, rounded
c  according to the least significant word.
c  In this Fortran version I do it on the first call in a way that
c  should get close to machine accuracy on any machine.
c  Variable t takes values tan(pi/4), tan(pi/8).. generated by solving
c  tan(theta) = 2*tan(theta/2)/(1 - tan(theta/2)**2) for tan(theta/2).
c  Then use sin(theta) = 2*tan(theta/2)/(1 + tan(theta/2)**2).
      data sines(1)/0./
      if( sines(1).eq.0. ) then
       t = 1.
       do 1 is = 1,log2nmax
	sines(is) = (t + t)/(1. + t*t)
    1	t = t/(1. + sqrt(1. + t*t))
      endif

      cpft = 0
      if( n.eq.1 ) return

c  Set up various indices.
c
      inc = incp
      sgn = signp
      ninc = n*inc
      span = ninc
c  Ugly way to determine what power of two n is:
      it = 2
      do 3 is = 1,log2nmax
       if( it.ge.n ) go to 2
    3  it = it + it

c  Oops.  n must be too large.
      cpft = 1
      return

2     if ( it.ne.n ) then
c  Oops.  n is not a power of 2.
        cpft = 2
        return
      endif
       
      go to 12

c    There are 2 inner loops which run over the n/(2*span) replications
c  of transforms of length (2*span).  These loops fit into the
c  instruction stack of the 6600 or 7600.  One loop is for arbitrary
c  rotation factor angle.  The other takes care of the special case in
c  which the angle is zero so that no complex multiplication is needed.
c  This is more efficient than testing and branching inside the inner
c  loop, as is often done.  The other special case in which no complex
c  multiply is needed is angle=pi (i.e. factor=i); this is not handled
c  specially.  These measures are most helpful for small n.
c
c    The organization of the recursion is that of Sande (ref. (3),
c  pp. 566-568).  That is, the data is in normal order to start and
c  scrambled afterward, and the exponential rotation ('twiddle') factor
c  angles are used in ascending order during each recursion level.
c  All the sines and cosines needed are generated from a short table
c  using a stable multiple-angle recursion (ref. (1), p. 651 and ref. (2),
c  pp. 179-180).  This method is economical in storage and time, and
c  yields accuracy comparable to good library sin-cos routines.
c  Angles between 0 and pi are needed.  The recursion is used for
c  angles up to pi/2; larger angles are obtained by reflection in the
c  imaginary axis (angle:=pi-angle).  These pairs of angles are used
c  one right after the other.
c
c    For simplicity, commentary below applies to inc=1 case.
c
c  If truncated rather than rounded arithmetic is used, Singleton's
c  magnitude correction should be applied to c and s.
c
10    t = s + (s0*c - c0*s)
      c = c - (c0*c + s0*s)
      s = t
c  Replication loop.
11    k1 = k0 + span
      r0 = r(k0)
      r1 = r(k1)
      i0 = i(k0)
      i1 = i(k1)
      r(k0) = r0 + r1
      i(k0) = i0 + i1
      r0 = r0 - r1
      i0 = i0 - i1
      r(k1) = c*r0 - s*i0
      i(k1) = s*r0 + c*i0
      k0 = k1 + span
      if( k0.lt.ninc ) go to 11
      k1 = k0 - ninc
      c = -c
      k0 = span - k1
      if( k1.lt.k0 ) go to 11
      k0 = k0 + inc
      if( k0.lt.k1 ) go to 10
c  Recursion to next level.
12    continue
      span = span/2
      k0 = 0
c  Angle=0 loop.
13    k1 = k0 + span
      r0 = r(k0)
      r1 = r(k1)
      i0 = i(k0)
      i1 = i(k1)
      r(k0) = r0 + r1
      i(k0) = i0 + i1
      r(k1) = r0 - r1
      i(k1) = i0 - i1
      k0 = k1 + span
      if( k0.lt.ninc ) go to 13
c  Are we finished...
      if( span.eq.inc ) go to 20
c  No.  Prepare non-zero angles.
      c0 = 2.*sines(is)**2
      is = is - 1
      s = sign( sines(is),sgn )
      s0 = s
      c = 1. - c0
      k0 = inc
      go to 11
c
c    Arrays r and i now contain transform, but stored in 'reverse-
c    binary' order.  The re-ordering is done by pair exchanges.
c    Reference for sorting principle is p. 180 and p. 182 of ref. (2).
c
c    Once again, commentary applies to inc=1 case.
c  Indices are:
c    ij = 0,1,2...n/2-1 ( a simple counter).
c    ji = reversal of ij.
c    rc = reversal of 0,2,4...n/2 (incremented n/4 times).
c  rc is incremented thusly: starting with the next-to-leftmost bit,
c  change each bit up to and including first 0.  (The actual coding is
c  done so as to work for any inc>0 with equal efficiency.)
c    For all exchanges ij fits one of these cases:
c      (1) 1st and last bits are 0 (ij,ji even and <n/2), and ij<=ji.
c      (2) one's complement of case (1) (both odd and >n/2).
c      (3) 1st bit 0, last bit 1 (ij odd and <n/2, ji>n/2).
c    The code from label even down to odd is entered with ij even and
c  <=ji.  First time thru the complements are done -case (2).  Second
c  time thru gets case (1).  Thus a pair of elements both in the first
c  half of the sequence, and another pair in the 2nd half, are
c  exchanged.  The condition ij<ji prevents a pair from being exchanged
c  twice.
c    The code from label odd down to increv does case (3).

20    n1 = ninc - inc
      n2 = ninc/2
      rc = 0
      ij = 0
      ji = 0
      if( n2.eq.inc ) return
      go to 22

c  Even.
21    ij = n1 - ij
      ji = n1 - ji
      t = r(ij)
      r(ij) = r(ji)
      r(ji) = t
      t = i(ij)
      i(ij) = i(ji)
      i(ji) = t
      if( ij.gt.n2 ) go to 21
c  Odd.
22    ij = ij + inc
      ji = ji + n2
      t = r(ij)
      r(ij) = r(ji)
      r(ji) = t
      t = i(ij)
      i(ij) = i(ji)
      i(ji) = t
      it = n2
c  Increment reversed counter.
23    it = it/2
      rc = rc - it
      if( rc.ge.0 ) go to 23
      rc = rc + 2*it
      ji = rc
      ij = ij + inc
      if( ij.lt.ji ) go to 21
      if( ij.lt.n2 ) go to 22

      return
      end
